from peaq.utils import ExtrinsicBatch


def _rpc_id(entity_id):
    return [int(entity_id[i:i + 2], 16) for i in range(0, len(entity_id), 2)]


def _from_rpc_id(entity_id):
    return ''.join(format(i, '02x') for i in entity_id)


def _comp_rbac_call(batch, cl_fcn, cl_par):
    batch.compose_call(
        'PeaqRbac',
        cl_fcn,
        cl_par
    )


def rbac_add_role_payload(batch, entity_id, name):
    _comp_rbac_call(
        batch,
        'add_role',
        {
            'role_id': entity_id,
            'name': name,
        })


def rbac_add_role(substrate, kp_src, entity_id, name):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_add_role_payload(batch, entity_id, name)
    return batch.execute()


def rbac_add_group_payload(batch, group_id, name):
    _comp_rbac_call(
        batch,
        'add_group',
        {
            'group_id': group_id,
            'name': name,
        })


def rbac_add_group(substrate, kp_src, group_id, name):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_add_group_payload(batch, group_id, name)
    return batch.execute()


def rbac_add_permission_payload(batch, permission_id, name):
    _comp_rbac_call(
        batch,
        'add_permission',
        {
            'permission_id': permission_id,
            'name': name,
        })


def rbac_add_permission(substrate, kp_src, permission_id, name):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_add_permission_payload(batch, permission_id, name)
    return batch.execute()


def rbac_permission2role_payload(batch, permission_id, role_id):
    _comp_rbac_call(
        batch,
        'assign_permission_to_role',
        {
            'permission_id': permission_id,
            'role_id': role_id,
        })


def rbac_permission2role(substrate, kp_src, permission_id, role_id):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_permission2role_payload(batch, permission_id, role_id)
    return batch.execute()


def rbac_role2group_payload(batch, role_id, group_id):
    _comp_rbac_call(
        batch,
        'assign_role_to_group',
        {
            'role_id': role_id,
            'group_id': group_id,
        })


def rbac_role2group(substrate, kp_src, role_id, group_id):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_role2group_payload(batch, role_id, group_id)
    return batch.execute()


def rbac_role2user_payload(batch, role_id, user_id):
    _comp_rbac_call(
        batch,
        'assign_role_to_user',
        {
            'role_id': role_id,
            'user_id': user_id,
        })


def rbac_role2user(substrate, kp_src, role_id, user_id):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_role2user_payload(batch, role_id, user_id)
    return batch.execute()


def rbac_user2group_payload(batch, user_id, group_id):
    _comp_rbac_call(
        batch,
        'assign_user_to_group',
        {
            'user_id': user_id,
            'group_id': group_id,
        })


def rbac_user2group(substrate, kp_src, user_id, group_id):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_user2group_payload(batch, user_id, group_id)
    return batch.execute()


def rbac_disable_group_payload(batch, group_id):
    _comp_rbac_call(
        batch,
        'disable_group',
        {
            'group_id': group_id,
        })


def rbac_disable_group(substrate, kp_src, group_id):
    batch = ExtrinsicBatch(substrate, kp_src)
    rbac_disable_group_payload(batch, group_id)
    return batch.execute()


def _convert_output(data):
    for convert_type in ['id', 'name', 'role', 'group', 'permission', 'user']:
        if convert_type in data:
            if not data[convert_type]:
                continue
            if isinstance(data[convert_type][0], list):
                data[convert_type] = [bytes(item).decode('utf-8')
                                      for item in data[convert_type]]
                continue
            try:
                data[convert_type] = bytes(data[convert_type]).decode('utf-8')
            except UnicodeDecodeError:
                data[convert_type] = _from_rpc_id(data[convert_type])


def _rbac_rpc_fetch_entity(substrate, addr, entity, params):
    bl_hsh = substrate.get_block_hash(None)
    data = substrate.rpc_request(
        f'peaqrbac_fetch{entity}',
        [addr] + params + [bl_hsh]
    )['result']

    if 'Err' in data:
        data['Err']['param'] = _from_rpc_id(data['Err']['param'])
    elif 'Ok' in data:
        ok = data['Ok']
        if isinstance(ok, list):
            for item in ok:
                _convert_output(item)
        else:
            _convert_output(ok)
    return data


def rbac_rpc_fetch_role(substrate, addr, entity_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Role', [_rpc_id(entity_id)])


def rbac_rpc_fetch_permission(substrate, addr, entity_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Permission', [_rpc_id(entity_id)])


def rbac_rpc_fetch_group(substrate, addr, entity_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Group', [_rpc_id(entity_id)])


def rbac_rpc_fetch_roles(substrate, addr):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Roles', [])


def rbac_rpc_fetch_permissions(substrate, addr):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Permissions', [])


def rbac_rpc_fetch_groups(substrate, addr):
    return _rbac_rpc_fetch_entity(substrate, addr, 'Groups', [])


def rbac_rpc_fetch_group_roles(substrate, addr, group_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'GroupRoles', [_rpc_id(group_id)])


def rbac_rpc_fetch_group_permissions(substrate, addr, group_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'GroupPermissions', [_rpc_id(group_id)])


def rbac_rpc_fetch_role_permissions(substrate, addr, role_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'RolePermissions', [_rpc_id(role_id)])


def rbac_rpc_fetch_user_roles(substrate, addr, user_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'UserRoles', [_rpc_id(user_id)])


def rbac_rpc_fetch_user_groups(substrate, addr, user_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'UserGroups', [_rpc_id(user_id)])


def rbac_rpc_fetch_user_permissions(substrate, addr, user_id):
    return _rbac_rpc_fetch_entity(substrate, addr, 'UserPermissions', [_rpc_id(user_id)])



import time
from dataclasses import dataclass
from substrateinterface import SubstrateInterface, Keypair
from scalecodec.base import RuntimeConfiguration
from scalecodec.type_registry import load_type_registry_preset
from scalecodec.utils.ss58 import ss58_encode

DEBUG = False


def get_account_balance(substrate, addr, block_hash=None):
    result = substrate.query(
        'System', 'Account', [addr], block_hash=block_hash)
    return int(result['data']['free'].value)


def show_extrinsic(receipt, info_type):
    if receipt.is_success:
        print(f'🚀 {info_type}, Success: {receipt.get_extrinsic_identifier()}')
    else:
        print(f'💥 {info_type}, Extrinsic Failed: {receipt.error_message} {receipt.get_extrinsic_identifier()}')


def _generate_call_description(call):
    """Generates a description for an arbitrary extrinsic call"""
    # print(type(call), call)
    # assert type(call) == "scalecodec.types.GenericCall"
    module = call.call_module.name
    function = call.call_function.name
    if module == 'Sudo':
        # I don't like this solution, but unfortunately I was not able to access
        # call.call_args in that way to extract the module and function of the payload.
        desc = call.__str__().split('{')[3]
        desc = desc.split("'")
        submodule = desc[3]
        subfunction = desc[7]
        return f'{module}.{function}({submodule}.{subfunction})'
    else:
        return f'{module}.{function}'


def _generate_batch_description(batch):
    """Generates a description for an extrinsic batch"""
    desc = [f'{_generate_call_description(b)}' for b in batch]
    desc = ', '.join(desc)
    return f'Batch[ {desc} ]'


def into_keypair(keypair_or_uri) -> Keypair:
    """Takes either a Keypair, or transforms a given uri into one"""
    if isinstance(keypair_or_uri, str):
        return Keypair.create_from_uri(keypair_or_uri)
    elif isinstance(keypair_or_uri, Keypair):
        return keypair_or_uri
    else:
        raise TypeError


def into_substrate(substrate_or_url) -> SubstrateInterface:
    """Takes a SubstrateInterface, or takes into one by given url"""
    if isinstance(substrate_or_url, str):
        return SubstrateInterface(substrate_or_url)
    elif isinstance(substrate_or_url, SubstrateInterface):
        return substrate_or_url
    else:
        raise TypeError


@dataclass
class ExtrinsicBatch:
    """
    ExtrinsicBatch class for simple creation of extrinsic-batch to be executed.

    When initialising, pass either an existing SubstrateInterface/WS-URL and
    optional Keypair/URI, or use the defaults. The ExtrinsicBatch is designed
    to be used on one chain (relaychain/parachain), because the usage of one
    SubstrateInterface. It is also designed for one user to execute the batch,
    because the Utility pallet does not varying users unfortunately.

    Example 1:    ex_stack = ExtrinsicStack(substrate, kp_src)
    Example 2:    ex_stack = ExtrinsicStack(WS_URL, '//Bob')
    Example 3:    ex_stack = ExtrinsicStack()
    """
    substrate: SubstrateInterface
    keypair: Keypair
    batch: list

    def __init__(self, substrate_or_url, keypair_or_uri):
        self.substrate = into_substrate(substrate_or_url)
        self.keypair = into_keypair(keypair_or_uri)
        self.batch = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self

    def __str__(self):
        return 'ExtrinsicBatch@{}, batch: {}'.format(self.substrate, self.batch)

    def compose_call(self, module, extrinsic, params):
        """Composes and appends an extrinsic call to this stack"""
        self.batch.append(self._compose_call(
            self.substrate, module, extrinsic, params))

    def compose_sudo_call(self, module, extrinsic, params):
        """Composes a sudo-user extrinsic call and adds it this stack"""
        self.batch.append(self._compose_sudo_call(
            self.substrate, module, extrinsic, params))

    # TODO
    def execute(self, wait_for_finalization=False, alt_keypair=None) -> str:
        """Executes the extrinsic-stack"""
        if not self.batch:
            return ''
        if alt_keypair is None:
            alt_keypair = self.keypair
        return self._execute_extrinsic_batch(
            self.substrate, alt_keypair, self.batch, wait_for_finalization)

    # TODO
    def execute_n_clear(self, alt_keypair=None, wait_for_finalization=False) -> str:
        """Combination of execute() and clear()"""
        if alt_keypair is None:
            alt_keypair = self.keypair
        receipt = self.execute(wait_for_finalization, alt_keypair)
        self.clear()
        return receipt

    def clear(self):
        """Clears the current extrinsic-stack"""
        self.batch = []

    def clone(self, keypair_or_uri=None):
        """Creates a duplicate, by using the same SubstrateInterface"""
        if keypair_or_uri is None:
            keypair_or_uri = self.keypair
        return ExtrinsicBatch(self.substrate, keypair_or_uri)

    def get_calls(self):
        """returns a list of the alls of the extrinsics in this stack"""
        return [x for x in self.batch]

    def _compose_call(self, substrate, module, extrinsic, params):
        """
        Composes a substrate-extrinsic-call on any module
        Example:
          module = 'Rbac'
          extrinsic = 'add_role'
          params = {'role_id': entity_id, 'name': name }
        """
        return substrate.compose_call(
            call_module=module,
            call_function=extrinsic,
            call_params=params
        )

    def _compose_sudo_call(self, substrate, module, extrinsic, params):
        """
        Composes a substrate-sudo-extrinsic-call on any module
        Parameters same as in compose_call, see above
        """
        payload = self._compose_call(substrate, module, extrinsic, params)
        return self._compose_call(substrate, 'Sudo', 'sudo', {'call': payload.value})

    def _execute_extrinsic_batch(self, substrate, kp_src, batch,
                                 wait_for_finalization=False) -> str:
        """
        Executes a extrinsic-stack/batch-call on substrate
        Parameters:
          substrate:  SubstrateInterface
          kp_src:     Keypair
          batch:      list[_compose_call(), _compose_call(), ...]
        """
        # Wrap payload into a utility batch cal
        call = substrate.compose_call(
            call_module='Utility',
            call_function='batch_all',
            call_params={
                'calls': batch,
            })

        nonce = substrate.get_account_nonce(kp_src.ss58_address)
        extrinsic = substrate.create_signed_extrinsic(
            call=call,
            keypair=kp_src,
            era={'period': 64},
            nonce=nonce
        )

        receipt = substrate.submit_extrinsic(
            extrinsic, wait_for_inclusion=True,
            wait_for_finalization=wait_for_finalization)
        if len(batch) == 1:
            description = _generate_call_description(batch[0])
        else:
            description = _generate_batch_description(batch)
        if DEBUG:
            show_extrinsic(receipt, description)

        return receipt


def get_block_height(substrate):
    latest_block = substrate.get_block()
    return latest_block['header']['number']


def get_block_hash(substrate, block_num):
    return substrate.get_block_hash(block_id=block_num)


def get_chain(substrate):
    return substrate.rpc_request(method='system_chain', params=[]).get('result')


def wait_for_n_blocks(substrate, n=1):
    """Waits until the next block has been created"""
    height = get_block_height(substrate)
    wait_height = height + n
    past = 0
    while past < n:
        next_height = get_block_height(substrate)
        if height == next_height:
            time.sleep(1)
        else:
            print(f'Current block: {height}, but waiting at {wait_height}')
            height = next_height
            past = past + 1


def calculate_multi_sig(kps, threshold):
    '''https://github.com/polkascan/py-scale-codec/blob/f063cfd47c836895886697e7d7112cbc4e7514b3/test/test_scale_types.py#L383'''  # noqa: E501

    addrs = [kp.ss58_address for kp in kps]
    RuntimeConfiguration().update_type_registry(load_type_registry_preset('legacy'))
    multi_account_id = RuntimeConfiguration().get_decoder_class('MultiAccountId')

    multi_sig_account = multi_account_id.create_from_account_list(addrs, threshold)
    return ss58_encode(multi_sig_account.value.replace('0x', ''), 42)



from peaq.utils import ExtrinsicBatch


def did_add_payload(batch, did_account, name, value, valid_for=None):
    batch.compose_call('PeaqDid', 'add_attribute', {
        'did_account': did_account,
        'name': name,
        'value': value,
        'valid_for': valid_for,
    })


def did_add(substrate, kp_src, did_account, name, value, valid_for=None):
    batch = ExtrinsicBatch(substrate, kp_src)
    did_add_payload(batch, did_account, name, value, valid_for)
    return batch.execute()


def did_update_payload(batch, did_account, name, value, valid_for=None):
    batch.compose_call(
        'PeaqDid',
        'update_attribute',
        {
            'did_account': did_account,
            'name': name,
            'value': value,
            'valid_for': valid_for,
        })


def did_update(substrate, kp_src, did_account, name, value, valid_for=None):
    batch = ExtrinsicBatch(substrate, kp_src)
    did_update_payload(batch, did_account, name, value, valid_for)
    return batch.execute()


def did_remove_payload(batch, did_account, name):
    batch.compose_call(
        'PeaqDid',
        'remove_attribute',
        {
            'did_account': did_account,
            'name': name,
        })


def did_remove(substrate, kp_src, did_account, name):
    batch = ExtrinsicBatch(substrate, kp_src)
    did_remove_payload(batch, did_account, name)
    return batch.execute()


def did_rpc_read(substrate, did_account, name):
    bl_hsh = substrate.get_block_hash(None)
    data = substrate.rpc_request('peaqdid_readAttribute', [did_account, name, bl_hsh])
    return data['result']



# [TODO] Change API
def transfer_with_tip(substrate, kp_src, kp_dst_addr, token_num, tip, token_base=0):
    if not token_base:
        token_base = 10 ** 3

    nonce = substrate.get_account_nonce(kp_src.ss58_address)

    call = substrate.compose_call(
        call_module='Balances',
        call_function='transfer_keep_alive',
        call_params={
            'dest': kp_dst_addr,
            'value': token_num * token_base
        })

    extrinsic = substrate.create_signed_extrinsic(
        call=call,
        keypair=kp_src,
        era={'period': 64},
        tip=tip * token_base,
        nonce=nonce
    )

    receipt = substrate.submit_extrinsic(extrinsic, wait_for_inclusion=True)
    return receipt


# TODO Change API
def transfer(substrate, kp_src, kp_dst_addr, token_num, token_base=0):
    return transfer_with_tip(substrate, kp_src, kp_dst_addr, token_num, 0, token_base)



from setuptools import setup, find_packages

VERSION = '0.2.0'

setup(
    name='peaq-py',
    version=VERSION,
    url='https://github.com/peaqnetwork/peaq-py/',
    author='Jay Pan',
    author_email='pan@eotlabs.io',
    description='peaq python sdk',
    packages=find_packages(),
    install_requires=[],
)
